#include "JVizClientObj.h"
#include "JConvertObj.h"
#include "JLinkObj.h"
#include "JInteger.h"
#include "JIntegerData.h"
#include "JBlockData.h"
#include "JBytesData.h"
#include "JCharacter.h"
#include "JStringData.h"
#include "JStringProperty.h"
#include "JIntegerProperty.h"
#include "JSocketException.h"
#include "JUnknownHostException.h"
#include "VizCommands.h"
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <time.h>
char* newstrcpy( char* x ) {
	char* xx = (char*)malloc( strlen( x ) + 1 );
	strcpy( xx, x );
	return( xx );
}

static char* q[1000];
static int qFirst = 0;
static int qLast = 0;
static bool qFull = false;
static void incQfirst() {
	if ( qFirst != qLast ) {
		qFirst++;
	}
	if ( qFirst == 1000 ) {
		qFirst = 0;
	}
	if ( qFirst != qLast ) {
		qFull = false;
	}
}
static void incQlast() {
	if ( qFull ) {
		return;
	}
	qLast++;
	if ( qLast == 1000 ) {
		qLast = 0;
	}
	if ( qFirst == qLast ) {
		qFull = true;
	}
}
static bool qIsEmpty() {
	if ( !qFull && ( qFirst == qLast )) {
		return( true );
	} else {
		return( false );
	}
}

// queue is empty when qFirst==qLast
// q has no more room if incrementing qFirst makes qFirst==qLast
static void qPut( char* buf ) {
	if ( q[qLast] != NULL ) {
		free( q[qLast] );
	}
	q[qLast] = newstrcpy( buf );
	incQlast();
}

static char* qPop() {
	char* result = q[qFirst];
	incQfirst();
	return( result );
}


static void qPutObject( char* object ) {
	char buf[100];
	sprintf( buf, "setobject %s\n", object );
	qPut( &buf[0] );
}
static void qPutProperty( char* property ) {
	char buf[100];
	sprintf( buf, "setproperty %s\n", property );
	qPut( &buf[0] );
}
static void qPutValue( char* value ) {
	char buf[100];
	sprintf( buf, "setvalue %s\n", value );
	qPut( &buf[0] );
}

char* theJVizClient = JVizClient().Register();		  
void JVizClient::writeContent(JOutputStream& os) {
  JAddObj::writeContent(os);
  putObject(os, "address", address);
  putInteger(os, "port", port);
}

void JVizClient::readContent(JDictionary& dict) {
  JAddObj::readContent(dict);
  JObject *obj = getObject(dict, "address");
  if (obj) address = *(JString*)obj;
  port = getInteger(dict, "port");
}

const char* JVizClient::className() const { return "JVizClient";}
JObject* JVizClient::clone() const { 
//    JComponent::debug( "Clone JVizClient" );
    return new JVizClient(*this);
}

JVizClient::JVizClient() {
  address = "localhost";
  port = 3502;
  connected = false;
  enabled = false;
  ssequence = 1;
  igm = 3;
  ogm = 1;
  for ( int i = 0; i < 1000; i++ ) {
	  q[i] = NULL;
  }
}

JVizClient::JVizClient(const JVizClient& obj) { *this=obj;}
JVizClient::~JVizClient() { close();}


void JVizClient::startup() {
//    sendFileLoadRequest = true;
}

void JVizClient::reset() { 
  connected = false;
  data.Dereference();
  socket.close();
}



/** Get the output JStringData prepared to go outside the module */
void JVizClient::access(int n, JLinkObj& link, const JDataType& _data) { 
    JSTRING(_data) = result; 
    waitUntilBroadcasted = false; 
}


/**
 *  Module characteristics associated with inputs.
 */

/**
 *  inputType
 *  
 *  Purpose:  
 *     get the input type corresponding to location
 *     All input types are int except for DEF list, which is a string
 */

int JVizClient::inputType(int n) { 
    if ( n == 0) return JStringData::id;
    else return JIntegerData::id;
}


/**
 *  inputTag
 *
 *  Purpose:
 *     provide a JString describing an input.
 */

JString JVizClient::inputTag(int n) {
  char* tag[] = {
      // left inputs
        "command",            // 0

      // top inputs
        "enable(logical)",  // 9
  };
  return( tag[n] );
}


/**
 *  inputAllowed
 *
 *  Purpose:
 *     determines whether a new connection can be made to an input.
 *     For this module, multiple bytes connections are allowed, but only 
 *     a single "enable" connection is allowed.
 */

boolean JVizClient::inputAllowed(int n) { 
    return true; // allow as many inputs as possible -- was: !inputSet(n).last();
}


/** 
 *  inputArea
 *
 *  Purpose:
 *     Defines area within module associated with each input.
 *     x,y,z,signals 1-5, and transformer data are on left
 *     enable, timer sampling, calibration, and send calibration are on top
 */

JFRect JVizClient::inputArea(int n) {
  if (n < ENABLED) return leftArea(n, 0, 1);
  return topArea(n, ENABLED, 1);
}



/**
 *  inputFace
 *
 *  Purpose:
 *     Indicates which face on the module rectangle is associated with an input.
 */

int JVizClient::inputFace(int n) {
  if (n < ENABLED) return LEFT;
  return TOP;
}


/**
 *  inputPoint
 *
 *  Purpose:
 *     Given an input, provides the exact location of the connection
 *     to the module for that input.
 */

JFPoint JVizClient::inputPoint(int n, JLinkObj& link) {
  if (n < ENABLED) return leftPoint(n, link, 0, ENABLED);
  return topPoint(n, link, ENABLED, 1);
}


/**
 *  Module characteristics associated with outputs
 */

/**
 *  outputTag
 *
 *  Purpose:
 *      Provide a string description of an output.  This module has a
 *      single output which is a list of DEF names in the WRL file
 *      that was loaded into viz.
 */

JString JVizClient::outputTag(int n) { 
    return( "viz command" ); 
}


/**
 *  outputType
 *
 *  Purpose:  
 *      Provide a data type for an output.  This module has a single output
 *      which is a string.
 */

int JVizClient::outputType(int n) { 
    return JStringData::id;
} 

JFRect JVizClient::outputArea(int n) { 
    return rightArea(n, 0, 1, 0, 1);
}



JFPoint JVizClient::outputPoint(int n, JLinkObj& link) { 
    return rightPoint(n, link, 0, 1, 0, 1);
}

void JVizClient::paint(JGraphics g, double dx, double dy, JRegion& rgn, double scale) {
  JAddObj::paint(g, dx, dy, rgn, scale);
  JRect rect = getIExtent(dx, dy, scale);
  int kw = depth2*3;
  if ((rect.width > kw) && (rect.height > kw)) {
    if (connected > true) g.setJColor(JColor::red);
    else if (connected) g.setJColor(JColor::yellow);
    else g.setJColor(JColor::red.darker());
    g.fillRect(rect.x+depth2, rect.y+depth2, depth2, depth2);
  }
}

void JVizClient::draw(JGraphics g, int x, int y, int w, int h) {
  int dh = h/3;
  int dy = h/12;
  JRect rect1(x, y+dy, w, dh);
  JRect rect2(x, y+h/3, w, dh);
  JRect rect3(x, y+h*2/3-dy, w, dh);
  drawText(g, "Viz", rect1);
  drawText(g, "Client", rect2);
  drawText(g, address+":"+JInteger::toJString(port), rect3);
}




/**
 *  engine
 *
 *  Purpose:
 *     The main processing engine for this module, called whenever
 *     data is received on an incoming visual link.
 */
static char last_object[100];
static char last_property[100];
static char last_value[100];
void JVizClient::engine(int n, JLinkObj& link) {
    switch (n) {
    case COMMAND: {
		JString inCommand = command;
		link.access(JStringData(inCommand));
		if ( inCommand == command ) {
			break;
		} else {
			command = inCommand;
		}
				  }
        break;
    case ENABLED:
        link.access(JIntegerData(enabled));
        if (enabled && !connected) open();
        else if (!enabled && connected) {
			last_object[0] = 0;
			last_property[0] = 0;
			last_value[0] = 0;
			close();
		}
        break;
    default:
        break;
    }
	if ( enabled && connected ) {
		char* scanner = (char*)command;
		char* objectName = scanner;
		while ( *scanner && (*scanner != ',' )) {
			scanner++;
		}
		if ( *scanner == ',' ) {
			*scanner = 0;
			scanner++;
		}
		char* propertyName = scanner;
		while ( *scanner && (*scanner != ',' )) {
			scanner++;
		}
		if ( *scanner == ',' ) {
			*scanner = 0;
			scanner++;
		}
		char* value = scanner;
		if (( objectName != propertyName ) && ( propertyName != value )) {
			if ( strcmp( last_object, objectName ) != 0 ) {
				qPutObject( objectName );
				//sos << JBlock( JString( "setobject " ) + JString( objectName ) + JString( "\n" ));
				strcpy( last_object, objectName );
			}

			if ( strcmp( last_property, propertyName ) != 0 ) {
				qPutProperty( propertyName );
				//sos << JBlock( JString( "setproperty " ) + JString( propertyName ) + JString( "\n" ));
				strcpy( last_property, propertyName );
			}

			if ( strcmp( last_value, value ) != 0 ) {
				qPutValue( value );
				//sos << JBlock( JString( "setvalue " ) + JString( value ) + JString( "\n" ));
				strcpy( last_value, value );
			}
			*(propertyName - 1) = ',';
			*(value - 1) = ',';
		}
	} 
}


JArray JVizClient::getProperties() {
  JArray properties = JAddObj::getProperties();
  properties.append(JStringProperty("hostName/IPaddress", address));
  properties.append(JIntegerProperty("port", port, 1024, 5000));
  return properties;
}

boolean JVizClient::updateProperty(JProperty& prop) {
  if (JAddObj::updateProperty(prop)) return true;
  boolean changed = false;
  boolean oldenabled = enabled;
  close();
  if (prop.getName() == JString("port")) {
    port = ((JIntegerProperty*)&prop)->value;
    changed = true;
  } else if (prop.getName() == JString("hostName/IPaddress")) {
    address = ((JStringProperty*)&prop)->value;
    changed = true;
  }
  if (oldenabled) open();
  if (changed) repaint();
  return changed;
}



/**
 *  ======== JVizClientObj::run ========
 *
 *  Purpose:
 *    Thread for connecting to remote socket.
 */
void JVizClient::run() {
	// initially we are not connected
	connected = false;
RETRY:
	try {
		openSocket();
	} catch (JException *e) {
		delete e;
		closeSocket();
		return;
	}
	connected = true;
	repaint();
	try {
	  // Keep trying to connect until we do
		while (connected) {
		  if (connectSocket()) break;
		  JThread::sleep(100);
		}
	} catch (JException *e) {
		delete e;
		closeSocket();
		if (connected) goto RETRY;
		repaint();
		return;
	}
	if ( connected ) {
		// Here connected is (true+1) for some reason
		// probably to make it show up as red (yellow means trying to connect)
		connected++;
		repaint();
		while ( connected ) {
			if ( qIsEmpty() ) {
				JThread::sleep(10);
			} else {
				try {
					sos << JString(qPop()); //JThread::sleep(1000);
				} catch( JException *e ) {
					delete e;
					connected = 0;
				}
			}
		}
	}
    //goto RETRY;
}

/**
 *  open
 *
 *  Purpose:
 *      Start a thread (if not already running) to connect to external program
 *      via sockets.
 */

void JVizClient::open() {
  cs.lock();
  if (!connected) {
    thread = JThread(this);
    thread.start();
    thread.setPriority(JThread::TNORM_PRIORITY+1);
  }
  cs.unlock();
}


/**
 *  close
 *
 *  Purpose:
 *      Stop the thread (if not already stopped) that is connected to an
 *      external program via sockets.
 */

void JVizClient::close() {
  cs.lock();
  if (connected) {
    connected = false;
	//thread.setPriority(JThread::TNORM_PRIORITY);
    //closeSocket();
	//thread.waitFor();
    //repaint();
    //ssequence = 1;
  }
  cs.unlock();
}

void JVizClient::openSocket() {
  boolean isIPAddr = true;
  for (int i=0; address[i]; i++) 
    if (JCharacter::isAlpha(address[i]))
      isIPAddr = false;
  if (isIPAddr) {
    JInetAddress inet = JInetAddress(
      JInetAddress::getAddrByJString(address));
    socket = JSocket(inet, port);
    address = inet.getAddressJString();
  } else {
    JInetAddress inet = JInetAddress(address);
    socket = JSocket(inet, port);
    address = inet.getHostName();
  }
}

boolean JVizClient::connectSocket() {
  if (socket.connect()) {
    sis = JSocketInputStream(socket);
    sos = JSocketOutputStream(socket);
    return true;
  }
  return false;
}

void JVizClient::closeSocket() {
  try {
    socket.close();
  } catch (JException *e) {
    delete e;
  }
}
