#include "JVizObj.h"
#include "JInteger.h"
#include "JIntegerData.h"
#include "JIntegerProperty.h"
#include "JIntegerListProperty.h"
#include "JStringProperty.h"
#include "JLinkObj.h"
#include "JBytesData.h"
#include "VizCommands.h"

char* theJVizObj = JVizObj().Register();



const char* JVizObj::className() const { return "JVizObj";}


JObject* JVizObj::clone() const { return new JVizObj(*this);}

int JVizObj::outputType(int n) { return JBytesData::id; }

// class constructor
JVizObj::JVizObj() {
  igm = 11;
  ogm = 1;
  timer_sampled = 0;
  x = y = z = signal1 = signal2 = signal3 = signal4 = signal5 = 0;
  calibrating = false;
  send_calibration = false;
  calibrated = false;
  calibration_data_sent = false;
  initial_X_values_set = false;
  initial_Y_values_set = false;
  initial_Z_values_set = false;
  initial_signal1_values_set = false;
  initial_signal2_values_set = false;
  initial_signal3_values_set = false;
  initial_signal4_values_set = false;
  initial_signal5_values_set = false;
}


void JVizObj::draw(JGraphics g, int x, int y, int w, int h) {
  JRect rect(x, y, w, h);
  if ( calibrating ) {
      g.setJColor( JColor::red );
  } else if ( calibrated && initial_X_values_set && initial_Y_values_set && initial_Z_values_set && 
      ( initial_signal1_values_set || initial_signal2_values_set || initial_signal3_values_set ||initial_signal4_values_set ||initial_signal5_values_set )) {
      g.setJColor( JColor::green.darker() );
  } else {
      g.setJColor( JColor::black );
  }
  drawText(g, JString("Viz"), rect);
}


boolean JVizObj::inputAllowed(int n)
{ return !inputSet(n).last();}
JString JVizObj::inputTag(int n) {
  char* tag[] = {
    "x",
    "y",
	"z",
	"signal 1",
    "signal 2",
    "signal 3",
    "signal 4",
    "signal 5",
    "timer_sampled",
    "calibrate",
    "send_calibration" };
  return( tag[n] );
}


JString JVizObj::outputTag(int n) {
  char* tag[] = {
    "data out" };
  return( tag[n] );
}


JFRect JVizObj::inputArea(int n) {
  if ( n < ( 8 )) {
    return leftArea( n, 0, 8, 0.1, 0.8 );
  } else {
    return topArea( n, 8, 3, 0.25 );
  }
}


JFRect JVizObj::outputArea(int n) {
  return rightArea(n, 0, 1);
}


int JVizObj::inputFace(int n) {
  if ( n < 8 ) {
    return LEFT;
  } else {
    return TOP;
  }
}


int JVizObj::outputFace(int n) {
  return RIGHT;
}


JFPoint JVizObj::inputPoint(int n, JLinkObj& link) {
  if (n<8) return leftPoint(n, link, 0, 8, 0.1, 0.8);
  return topPoint(n, link, 8, 3);
}


JFPoint JVizObj::outputPoint(int n, JLinkObj& link) {
  return rightPoint(n, link, 0, 1);
}


void JVizObj::engine(int n, JLinkObj& link) {
  switch (n) {
  case X:
    link.access(JIntegerData(x));
    if ( calibrating ) {
        if ( !initial_X_values_set ) {
            minX = maxX = x;
            initial_X_values_set = true;
        } else if ( x < minX ) {
            minX = x;
        } else if ( x > maxX ) {
            maxX = x;
        }
    }
    break;
  case Y:
    link.access(JIntegerData(y));
    if ( calibrating ) {
        if ( !initial_Y_values_set ) {
            minY = maxY = y;
            initial_Y_values_set = true;
        } else if ( y < minY ) {
            minY = y;
        } else if ( y > maxY ) {
            maxY = y;
        }
    }
    break;
  case Z:
    link.access(JIntegerData(z));
    if ( calibrating ) {
        if ( !initial_Z_values_set ) {
            minZ = maxZ = z;
            initial_Z_values_set = true;
        } else if ( z < minZ ) {
            minZ = z;
        } else if ( z > maxZ ) {
            maxZ = z;
        }
    }
    break;
  case SIGNAL1:
    link.access(JIntegerData(signal1));
    if ( calibrating ) {
        if ( !initial_signal1_values_set ) {
            minSignal1 = maxSignal1 = signal1;
            initial_signal1_values_set = true;
        } else if ( signal1 < minSignal1 ) {
            minSignal1= signal1;
        } else if ( signal1> maxSignal1 ) {
            maxSignal1 = signal1;
        }
    }
    break;
  case SIGNAL2:
    link.access(JIntegerData(signal2));
    if ( calibrating ) {
        if ( !initial_signal2_values_set ) {
            minSignal2 = maxSignal2 = signal2;
            initial_signal2_values_set = true;
        } else if ( signal2< minSignal2 ) {
            minSignal2 = signal2;
        } else if ( signal2 > maxSignal2 ) {
            maxSignal2 = signal2;
        }
    }
    break;
  case SIGNAL3:
    link.access(JIntegerData(signal3));
    if ( calibrating ) {
        if ( !initial_signal3_values_set ) {
            minSignal3 = maxSignal3 = signal3;
            initial_signal3_values_set = true;
        } else if ( signal3 < minSignal3 ) {
            minSignal3 = signal3;
        } else if ( signal3 > maxSignal3) {
            maxSignal3 = signal3;
        }
    }
    break;
  case SIGNAL4:
    link.access(JIntegerData(signal4));
    if ( calibrating ) {
        if ( !initial_signal4_values_set ) {
            minSignal4 = maxSignal4 = signal4;
            initial_signal4_values_set = true;
        } else if ( signal4 < minSignal4 ) {
            minSignal4 = signal4;
        } else if ( signal4 > maxSignal4 ) {
            maxSignal4 = signal4;
        }
    }
    break;
  case SIGNAL5:
    link.access(JIntegerData(signal5));
    if ( calibrating ) {
        if ( !initial_signal5_values_set ) {
            minSignal5 = maxSignal5 = signal5;
            initial_signal5_values_set = true;
        } else if ( signal5 < minSignal5 ) {
            minSignal5 = signal5;
        } else if ( signal5 > maxSignal5 ) {
            maxSignal5 = signal5;
        }
    }
    break;
  case TIMER_SAMPLED:
    link.access(JIntegerData(timer_sampled));
    break;
  case CALIBRATING:
    link.access(JIntegerData(calibrating));
    break;
  case SEND_CALIBRATION:
    link.access(JIntegerData(send_calibration));
    break;
  default:
    break;
  }
  if ( n == TIMER_SAMPLED ) {
      if ( timer_sampled ) {
          if ( calibrated ) {
		    broadcast( 0 );
          } 
	  }
  } else if ( n == CALIBRATING ) {
      if ( calibrating ) {
          startCalibration();
      } else {
          endCalibration();
      }
  } else if ( n == SEND_CALIBRATION ) {
      if ( send_calibration && !calibrating && initial_X_values_set && initial_Y_values_set && initial_Z_values_set && 
          ( initial_signal1_values_set || initial_signal2_values_set || initial_signal3_values_set|| initial_signal4_values_set|| initial_signal5_values_set )) {
          calibration_data_sent = false;
          broadcast( 0 );
      } 
  }
}

void JVizObj::startCalibration() {
    calibrating = true;
    calibrated = false;
    initial_X_values_set = false;
    initial_Y_values_set = false;
    initial_Z_values_set = false;
    initial_signal1_values_set = false;
    initial_signal2_values_set = false;
    initial_signal3_values_set = false;
    initial_signal4_values_set = false;
    initial_signal5_values_set = false;
}

void JVizObj::endCalibration() {
    calibrating = false;
    calibrated = true;
}

JArray JVizObj::getProperties() {
  JArray properties;
  return properties;
}


boolean JVizObj::updateProperty(JProperty& prop) {
  if (prop.getName() == JString("profileName")) {
    return true;
  }
  return false;
}


void JVizObj::access(int n, JLinkObj& link, const JDataType& data) { 
    if ( !calibration_data_sent ) {
        int command = MAIN_CALIBRATION;
        int numberSignals = 5;
	    dataOut = JBlock(command) + JBlock(minX) + JBlock(maxX) + JBlock(minY) + JBlock(maxY) + JBlock(minZ) + JBlock(maxZ) + JBlock( numberSignals) + JBlock(initial_signal1_values_set) + JBlock(minSignal1) + JBlock(maxSignal1) + JBlock(initial_signal2_values_set) + JBlock(minSignal2) + JBlock(maxSignal2) + JBlock(initial_signal3_values_set) + JBlock(minSignal3) + JBlock(maxSignal3) + JBlock(initial_signal4_values_set) + JBlock(minSignal4) + JBlock(maxSignal4) + JBlock(initial_signal5_values_set) + JBlock(minSignal5) + JBlock(maxSignal5);
        calibration_data_sent = true;
    } else {
        int command = MAIN_DATA;
        int numberSignals = 5;
	    dataOut = JBlock(command) + JBlock(x) + JBlock(y) + JBlock(z) + JBlock(numberSignals) + JBlock(signal1) + JBlock(signal2) + JBlock(signal3) + JBlock(signal4) + JBlock(signal5);
    }
    data.assign(dataOut);
}
